@@ -2,23 +2,29 @@ package Test2::Hub;
 use strict;
 use warnings;
 
-use Carp qw/carp croak/;
-use Test2::Hub::State();
+use Carp qw/carp croak confess/;
 use Test2::Util qw/get_tid/;
 
 use Scalar::Util qw/weaken/;
 
 use Test2::Util::HashBase qw{
     pid tid hid ipc
-    state
     no_ending
-    _todo _meta parent_todo
+    _meta
     _filters
     _listeners
     _follow_ups
     _formatter
     _context_init
     _context_release
+
+    count
+    failed
+    ended
+    bailed_out
+    _passing
+    _plan
+    skip_reason
 };
 
 my $ID_POSTFIX = 1;
@@ -29,10 +35,11 @@ sub init {
     $self->{+TID} = get_tid();
     $self->{+HID} = join '-', $self->{+PID}, $self->{+TID}, $ID_POSTFIX++;
 
-    $self->{+_TODO} = [];
     $self->{+_META} = {};
 
-    $self->{+STATE} ||= Test2::Hub::State->new;
+    $self->{+COUNT}    = 0;
+    $self->{+FAILED}   = 0;
+    $self->{+_PASSING} = 1;
 
     if (my $formatter = delete $self->{formatter}) {
         $self->format($formatter);
@@ -43,6 +50,19 @@ sub init {
     }
 }
 
+sub reset_state {
+    my $self = shift;
+
+    $self->{+COUNT} = 0;
+    $self->{+FAILED} = 0;
+    $self->{+_PASSING} = 1;
+
+    delete $self->{+_PLAN};
+    delete $self->{+ENDED};
+    delete $self->{+BAILED_OUT};
+    delete $self->{+SKIP_REASON};
+}
+
 sub inherit {
     my $self = shift;
     my ($from, %params) = @_;
@@ -65,17 +85,6 @@ sub inherit {
     }
 }
 
-sub _fast_todo {
-    my ($self) = @_;
-    my $array = $self->{+_TODO};
-    pop @$array while @$array && !defined $array->[-1];
-    my $todo = @$array ? ${$array->[-1]} : undef;
-    return (
-        diag_todo => defined($todo) || $self->{+PARENT_TODO},
-        todo      => $todo,
-    )
-}
-
 sub meta {
     my $self = shift;
     my ($key, $default) = @_;
@@ -101,34 +110,6 @@ sub delete_meta {
     delete $self->{+_META}->{$key};
 }
 
-sub set_todo {
-    my $self = shift;
-    my ($reason) = @_;
-
-    unless (defined wantarray) {
-        carp "set_todo(...) called in void context, todo not set!";
-        return;
-    }
-
-    unless(defined $reason) {
-        carp "set_todo() called with undefined argument, todo not set!";
-        return;
-    }
-
-    my $ref = \$reason;
-    push @{$self->{+_TODO}} => $ref;
-    weaken($self->{+_TODO}->[-1]);
-    return $ref;
-}
-
-sub get_todo {
-    my $self = shift;
-    my $array = $self->{+_TODO};
-    pop @$array while @$array && !defined($array->[-1]);
-    return undef unless @$array;
-    return ${$array->[-1]};
-}
-
 sub format {
     my $self = shift;
 
@@ -270,9 +251,17 @@ sub process {
         }
     }
 
-    my $state = $self->{+STATE};
-    $e->update_state($state);
-    my $count = $state->count;
+    my $type = ref($e);
+    my $is_ok = $type eq 'Test2::Event::Ok';
+    my $no_fail = $type eq 'Test2::Event::Diag' || $type eq 'Test2::Event::Note';
+    my $causes_fail = $is_ok ? !$e->{effective_pass} : $no_fail ? 0 : $e->causes_fail;
+
+    $self->{+COUNT}++  if $is_ok || (!$no_fail && $e->increments_count);
+    $self->{+FAILED}++ and $self->{+_PASSING} = 0 if $causes_fail;
+
+    my $callback = $e->callback($self) unless $is_ok || $no_fail;
+
+    my $count = $self->{+COUNT};
 
     $self->{+_FORMATTER}->write($e, $count) if $self->{+_FORMATTER};
 
@@ -280,6 +269,8 @@ sub process {
         $_->{code}->($self, $e, $count) for @{$self->{+_LISTENERS}};
     }
 
+    return $e if $is_ok || $no_fail;
+
     my $code = $e->terminate;
     $self->terminate($code, $e) if defined $code;
 
@@ -307,24 +298,23 @@ sub finalize {
     my ($trace, $do_plan) = @_;
 
     $self->cull();
-    my $state = $self->{+STATE};
 
-    my $plan   = $state->plan;
-    my $count  = $state->count;
-    my $failed = $state->failed;
+    my $plan   = $self->{+_PLAN};
+    my $count  = $self->{+COUNT};
+    my $failed = $self->{+FAILED};
 
     # return if NOTHING was done.
     return unless $do_plan || defined($plan) || $count || $failed;
 
-    unless ($state->ended) {
+    unless ($self->{+ENDED}) {
         if ($self->{+_FOLLOW_UPS}) {
             $_->($trace, $self) for reverse @{$self->{+_FOLLOW_UPS}};
         }
 
         # These need to be refreshed now
-        $plan   = $state->plan;
-        $count  = $state->count;
-        $failed = $state->failed;
+        $plan   = $self->{+_PLAN};
+        $count  = $self->{+COUNT};
+        $failed = $self->{+FAILED};
 
         if (($plan && $plan eq 'NO PLAN') || ($do_plan && !$plan)) {
             $self->send(
@@ -333,11 +323,85 @@ sub finalize {
                     max => $count,
                 )
             );
-            $plan = $state->plan;
         }
+        $plan = $self->{+_PLAN};
     }
 
-    $state->finish($trace->frame);
+    my $frame = $trace->frame;
+    if($self->{+ENDED}) {
+        my (undef, $ffile, $fline) = @{$self->{+ENDED}};
+        my (undef, $sfile, $sline) = @$frame;
+
+        die <<"        EOT"
+Test already ended!
+First End:  $ffile line $fline
+Second End: $sfile line $sline
+        EOT
+    }
+
+    $self->{+ENDED} = $frame;
+    $self->is_passing(); # Generate the final boolean.
+}
+
+sub is_passing {
+    my $self = shift;
+
+    ($self->{+_PASSING}) = @_ if @_;
+
+    # If we already failed just return 0.
+    my $pass = $self->{+_PASSING} || return 0;
+    return $self->{+_PASSING} = 0 if $self->{+FAILED};
+
+    my $count = $self->{+COUNT};
+    my $ended = $self->{+ENDED};
+    my $plan = $self->{+_PLAN};
+
+    return $pass if !$count && $plan && $plan =~ m/^SKIP$/;
+
+    return $self->{+_PASSING} = 0
+        if $ended && (!$count || !$plan);
+
+    return $pass unless $plan && $plan =~ m/^\d+$/;
+
+    if ($ended) {
+        return $self->{+_PASSING} = 0 if $count != $plan;
+    }
+    else {
+        return $self->{+_PASSING} = 0 if $count > $plan;
+    }
+
+    return $pass;
+}
+
+sub plan {
+    my $self = shift;
+
+    return $self->{+_PLAN} unless @_;
+
+    my ($plan) = @_;
+
+    confess "You cannot unset the plan"
+        unless defined $plan;
+
+    confess "You cannot change the plan"
+        if $self->{+_PLAN} && $self->{+_PLAN} !~ m/^NO PLAN$/;
+
+    confess "'$plan' is not a valid plan! Plan must be an integer greater than 0, 'NO PLAN', or 'SKIP'"
+        unless $plan =~ m/^(\d+|NO PLAN|SKIP)$/;
+
+    $self->{+_PLAN} = $plan;
+}
+
+sub check_plan {
+    my $self = shift;
+
+    return undef unless $self->{+ENDED};
+    my $plan = $self->{+_PLAN} || return undef;
+
+    return 1 if $plan !~ m/^\d+$/;
+
+    return 1 if $plan == $self->{+COUNT};
+    return 0;
 }
 
 sub DESTROY {
@@ -496,62 +560,6 @@ This will delete all data in the specified metadata key.
 
 This method will retrieve the value of any meta-data key specified.
 
-=item $string = $hub->get_todo()
-
-Get the current TODO reason. This will be undef if there is no active todo.
-Please note that 0 and C<''> (empty string) count as active todo.
-
-=item $ref = $hub->set_todo($reason)
-
-This will set the todo message. The todo will remain in effect until you let go
-of the reference returned by this method.
-
-    {
-        my $todo = $hub->set_todo("Broken");
-
-        # These ok events will be TODO
-        ok($foo->doit, "do it!");
-        ok($foo->doit, "do it again!");
-
-        # The todo setting goes away at the end of this scope.
-    }
-
-    # This result will not be TODO.
-    ok(1, "pass");
-
-You can also do it without the indentation:
-
-    my $todo = $hub->set_todo("Broken");
-
-    # These ok events will be TODO
-    ok($foo->doit, "do it!");
-    ok($foo->doit, "do it again!");
-
-    # Unset the todo
-    $todo = undef;
-
-    # This result will not be TODO.
-    ok(1, "pass");
-
-This method can be called while TODO is already in effect and it will work in a
-sane way:
-
-    {
-        my $first_todo = $hub->set_todo("Will fix soon");
-
-        ok(0, "Not fixed"); # TODO: Will fix soon
-
-        {
-            my $second_todo = $hub->set_todo("Will fix eventually");
-            ok(0, "Not fixed"); # TODO: Will fix eventually
-        }
-
-        ok(0, "Not fixed"); # TODO: Will fix soon
-    }
-
-This also works if you free todo's out of order. The most recently set todo
-that is still active will always be used as the todo.
-
 =item $old = $hub->format($formatter)
 
 Replace the existing formatter instance with a new one. Formatters must be
@@ -581,9 +589,8 @@ returned by the C<listen()> method.
 
 =item $hub->follow_op(sub { ... })
 
-Use this to add behaviors that are called just before the
-L<Test2::Hub::State> for the hub is finalized. The only argument to your
-codeblock will be a L<Test2::Util::Trace> instance.
+Use this to add behaviors that are called just before the hub is finalized. The
+only argument to your codeblock will be a L<Test2::Util::Trace> instance.
 
     $hub->follow_up(sub {
         my ($trace, $hub) = @_;
@@ -653,9 +660,49 @@ This can be used to disable auto-ending behavior for a hub. The auto-ending
 behavior is triggered by an end block and is used to cull IPC events, and
 output the final plan if the plan was 'no_plan'.
 
-=item $bool = $hub->parent_todo
+=back
 
-This will be true if this hub is a child hub who's parent had todo set.
+=head2 STATE METHODS
+
+=over 4
+
+=item $hub->reset_state()
+
+Reset all state to the start. This sets the test count to 0, clears the plan,
+removes the failures, etc.
+
+=item $num = $hub->count
+
+Get the number of tests that have been run.
+
+=item $num = $hub->failed
+
+Get the number of failures (Not all failures come from a test fail, so this
+number can be larger than the count).
+
+=item $bool = $hub->ended
+
+True if the testing has ended. This MAY return the stack frame of the tool that
+ended the test, but that is not guarenteed.
+
+=item $bool = $hub->is_passing
+
+=item $hub->is_passing($bool)
+
+Check if the overall test run is a failure. Can also be used to set the
+pass/fail status.
+
+=item $hub->plan($plan)
+
+=item $plan = $hub->plan
+
+Get or set the plan. The plan must be an integer larger than 0, the string
+'no_plan', or the string 'skip_all'.
+
+=item $bool = $hub->check_plan
+
+Check if the plan and counts match, but only if the tests have ended. If tests
+have not unded this will return undef, otherwise it will be a true/false.
 
 =back
 
